"""Importer for PDF statement from Hang Seng Bank in Hong Kong.

Depends on external library pdftotext, which in many OS is packaged under poppler
"""
__copyright__ = "Copyright (C) 2019 Cheong Yiu Fung"
__license__ = "GNU GPLv3"

import struct
import re
from beancount.ingest import importer
from beancount.core.amount import Amount
from beancount.core import data
from beancount.core import flags
from beancount.core.number import D
from datetime import datetime

from beancount_hangseng import utils


class HangSengSavingsImporter(importer.ImporterProtocol):
    """An importer for Hang Seng Bank PDF statements."""

    def __init__(self, account_filing, currency, *, unpack_format='11s58s35s25s24s', debug=False):
        self.account_filing = account_filing
        self.currency = currency
        self.unpack_format = unpack_format
        self.debug = debug
        self.pad_width = sum([int(x) for x in self.unpack_format.split('s')[:-1]])

    def identify(self, f):
        if f.mimetype() != 'application/pdf':
            return False
        # The name "HANG SENG BANK" is in the logo as image, use bank code to
        # identify instead, which is 024.
        text = f.convert(utils.pdf_to_text)
        if text:
            return re.search('Bank code +024', text) is not None

    def extract(self, f, existing_entries=None):
        text = f.convert(utils.pdf_to_text)
        # Each section of account begins with "Integrated Account Statement
        # Savings". Extract everything non-greedily (*?) until there's a page
        # break (\n\n\n), or when it ends with the row of "Transaction Summary"
        SAVINGS_REGEXP = "Integrated Account Statement Savings\n.*\n.*\n\n(?P<record>(.|\n)*?)(?=\n\n|Transaction Summary|Credit Interest Accrued)"
        allmatches = re.findall(SAVINGS_REGEXP, text)
        # For each match result, the 2nd group is the ending page break or
        # 'Transaction Summary', which we don't need.
        record_corpus = '\n'.join(match[0] for match in allmatches)
        return self.get_txns_from_text(record_corpus, f)

    def file_name(self, f):
        return "HangSeng_{}_{}.pdf".format(self.file_account(f), self.file_date(f).strftime("%Y%m%d"))

    def file_account(self, f):
        # Get account from eStatement
        text = f.convert(utils.pdf_to_text)
        match = re.search('Account Number +(.*)', text)
        if match:
            return match.group(1)

    def file_date(self, f):
        # Get statement date from eStatement
        text = f.convert(utils.pdf_to_text)
        match = re.search('Statement Date +(.*)', text)
        if match:
            return datetime.strptime(match.group(1), "%d %b %Y").date()

    def get_txns_from_text(self, corpus, f):
        statement_date = self.file_date(f)
        lines = corpus.split('\n')
        entries = []
        trans_title = ''  # Initialize title
        for line_no in range(len(lines)):
            # A heuristic unpack approach to get all fields. Strip spaces for
            # easier post-process.
            line = lines[line_no]
            post_date, title, deposit, withdraw, balance = [x.decode().strip() for x in struct.unpack(self.unpack_format, str.encode(line.ljust(self.pad_width)))]
            if title in ["B/F BALANCE", "C/F BALANCE"]:
                continue  # Skip the first and last row

            trans_title = ' '.join([trans_title, ' '.join(title.split())])

            if self.debug:
                print("{0: >10} {1: >30} Deposit: {2: >15} Withdraw: {3: >15}  Balance: {4: >15}".format(post_date, title, deposit, withdraw, balance))
            if post_date:  # update transaction date
                trans_date = datetime.strptime(post_date, '%d %b')
                # Cross-year handling
                if statement_date.month == 1 and trans_date.month == 12:
                    trans_date = trans_date.replace(year=statement_date.year - 1).date()
                else:
                    trans_date = trans_date.replace(year=statement_date.year).date()
            if deposit or withdraw:  # A new transaction
                trans_amount = D(deposit) if deposit else D('-' + withdraw)
                txn = data.Transaction(
                    meta=data.new_metadata(f.name, line_no),
                    payee=trans_title.strip(),
                    date=trans_date,
                    flag=flags.FLAG_OKAY,
                    narration="",
                    tags=set(),
                    links=set(),
                    postings=[],
                )
                txn.postings.append(
                    data.Posting(
                        account=self.account_filing,
                        units=Amount(trans_amount, self.currency),
                        cost=None,
                        price=None,
                        flag=None,
                        meta=None
                    )
                )
                entries.append(txn)
                trans_title = ''  # Reset title for next transaction
        return entries



"""Importer for PDF statement from DBS Credit Card in Hong Kong.

Depends on external library pdftotext, which in many OS is packaged under poppler
"""
__copyright__ = "Copyright (C) 2019 Cheong Yiu Fung"
__license__ = "GNU GPLv3"

import struct
import re
from beancount.ingest import importer
from beancount.core.amount import Amount
from beancount.core import data
from beancount.core import flags
from beancount.core.number import D
from datetime import datetime

from beancount_hangseng import utils


class DBSImporter(importer.ImporterProtocol):
    """An importer for DBS Card PDF statements."""

    def __init__(self, account_filing, currency, *, unpack_format='6s9s102s33s', debug=False):
        self.account_filing = account_filing
        self.currency = currency
        self.unpack_format = unpack_format
        self.debug = debug
        self.pad_width = sum([int(x) for x in self.unpack_format.split('s')[:-1]])

    def identify(self, f):
        if f.mimetype() != 'application/pdf':
            return False
        text = f.convert(utils.pdf_to_text)
        if text:
            return re.search('www.dbs.com', text) is not None

    def extract(self, f, existing_entries=None):
        text = f.convert(utils.pdf_to_text)
        # We only care about everything before GRAND TOTAL. There are some
        # transactions after that, but those are for next month.
        allmatches = re.findall("(?P<corpus>(.|\n)*?GRAND TOTAL)", text)
        text = '\n'.join(match[0] for match in allmatches)
        # print(text)
        # Each section of account begins with "TRANS DATE POST DATE" Row.
        # Extract everything non-greedily (*?) until there's a page break (Which
        # shows "SUMMARY OF ACTIVITY SINCE YOUR LAST STATEMENT", or "*****
        # FINANCE CHARGE RATES *****"
        SAVINGS_REGEXP = "([a-zA-Z] [0-9]{4}-[0-9]{4}-[0-9]{4}-[0-9]{4}|TRANS DATE *POST DATE).*(?P<record>(.|\n)*?)((?=GRAND TOTAL)|(?=[0-9]{5,}/[0-9]{5,}))"
        allmatches = re.findall(SAVINGS_REGEXP, text)
        # For each match result, the 1st group is start marker, while the 3rd group is end markers, which we don't need.
        record_corpus = '\n'.join(match[1] for match in allmatches)
        return self.get_txns_from_text(record_corpus, f)

    def file_name(self, f):
        return "DBS_{}_{}.pdf".format(self.file_account(f), self.file_date(f).strftime("%Y%m%d"))

    def file_account(self, f):
        # Get account from eStatement
        text = f.convert(utils.pdf_to_text)
        match = re.search('ACCOUNT NUMBER\s+([0-9]{4}-[0-9]{4}-[0-9]{4}-[0-9]{4})', text)
        if match:
            return match.group(1).replace(' ', '-')

    def file_date(self, f):
        text = f.convert(utils.pdf_to_text)
        match = re.search('STATEMENT DATE.*?([0-9]{2} (?:JAN|FEB|MAR|APR|MAY|JUN|JUL|AUG|SEP|OCT|NOV|DEC) [0-9]{4})', text)
        if match:
            return datetime.strptime(match.group(1), "%d %b %Y").date()

    def get_txns_from_text(self, corpus, f):
        """
BASIC CARD - CHEONG YIU FUNG 4518-3545-XXXX-XXXX
 22   SEP            23   SEP            7-ELEVEN, HK (1535)    SHATIN        HK                                                                     13.50
 24   SEP            25   SEP            THE H.K. MI-HOME       HONG KONG     HK                                                                    219.00
 24   SEP            26   SEP            MCDONALD'S-102-FULL WI HONG KONG     HK                                                                     53.50
 26   SEP            27   SEP            TSUI WAH RESTAURANT    MONG KOK      HK                                                                    119.00

        Observations:
        1) New transaction starts at lines with a new transaction date
        2) Amount is at the same line of new transaction
        """
        statement_date = self.file_date(f)
        def is_useful_lines(line):
            # Skip useless lines. It's either the OPENING BALANCE, or the line
            # that indicates beginning of transactions, which starts with
            # account number
            line = line.strip()
            return (not line.startswith("SUBTOTAL")) and (not line.startswith("ODD CENTS")) and (not line.startswith("PREVIOUS BALANCE")) and (not line.startswith("BASIC CARD"))
        lines = corpus.split('\n')
        lines = list(filter(is_useful_lines, lines))

        # Hmm. Next is a terrible trick here where we reconstruct the lines so
        # that dates could be aligned. See magic-number-master-power.png. (DBS
        # has similar issues)
        lines = [' '.join(l[:36].split())+l[36:] for l in lines]
        # Remove empty strings '' from list
        lines = list(filter(None, [l.rstrip() for l in lines]))
        if self.debug:
            print("Actual Parsed lines:")
            print('\n'.join(lines))
            print("padwidth: {}".format(self.pad_width))
            print("Account: {}".format(self.file_account(f)))
        # Prepare variables
        entries = []
        narration = ''  # Initialize narration
        for line_no in range(len(lines)):
            line = lines[line_no]
            if self.debug:
                print("Line: {}".format(line))

            # If starts with a digit, it indicates a date line so we parse it
            if line[0].isdigit():
                str_txn_date, str_post_date, description, str_amount = [x.decode().strip() for x in struct.unpack(self.unpack_format, str.encode(line.ljust(self.pad_width)))]
                if self.debug:
                    print("{0: >10} {1: >10} description {2: >20} amount {3: >15}".format(str_txn_date, str_post_date, description, str_amount))
                post_date = datetime.strptime(str_post_date, "%d %b")
                txn_date = datetime.strptime(str_txn_date, "%d %b")
                # Cross-year handling
                if statement_date.month == 1 and post_date.month == 12:
                    post_date = post_date.replace(year=statement_date.year - 1).date()
                    txn_date = txn_date.replace(year=statement_date.year - 1).date()
                else:
                    post_date = post_date.replace(year=statement_date.year).date()
                    txn_date = txn_date.replace(year=statement_date.year).date()
                amount = str_amount.replace(",", "")
                txn_amount = D(amount[:-2]) if amount[-2:] == 'CR' else -D(amount)
            
            if line[0].isdigit():
                # If it's a transaction line, description is extracted
                narration = ' '.join([narration, ' '.join(description.split())])
            else:
                # Otherwise, the whole line is description
                narration = ' '.join([narration, ' '.join(line.strip().split())])

            # Only create the new transaction when we see the next line starts
            # with a digit (indicating that's next transaction), or when we're
            # at the last line already.
            if line_no == (len(lines) - 1) or lines[line_no + 1][0].isdigit():
                entries.append(self.create_txn(f, line_no, narration.strip(), post_date, txn_date, txn_amount))
                narration = ''  # Reset title for next transaction
        return entries

    def create_txn(self, f, line_no, narration, post_date, txn_date, amount):
        txn = data.Transaction(
            meta=data.new_metadata(f.name, line_no, kvlist={'txn_date': txn_date}),
            payee="",
            date=post_date,
            flag=flags.FLAG_OKAY,
            narration=narration,
            tags=set(),
            links=set(),
            postings=[],
        )
        txn.postings.append(
            data.Posting(
                account=self.account_filing,
                units=Amount(amount, self.currency),
                cost=None,
                price=None,
                flag=None,
                meta=None
            )
        )
        return txn



from beancount_hangseng.utils import *
from beancount_hangseng.Savings import HangSengSavingsImporter
from beancount_hangseng.MPowerMasterCard import MPowerMasterImporter
from beancount_hangseng.DBS import DBSImporter



"""Parse HangSeng eStatement to CSV.
"""
__copyright__ = "Copyright (C) 2019 Cheong Yiu Fung"
__license__ = "GNU GPLv3"

import argparse
import csv
from os import path
import sys

from beancount.ingest.cache import _FileMemo
from beancount_hangseng.Savings import HangSengSavingsImporter
from beancount_hangseng.MPowerMasterCard import MPowerMasterImporter
from beancount_hangseng.DBS import DBSImporter


class CsvParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)


def main():
    parser = CsvParser()
    parser.add_argument('-o', '--output', default=None,
                        help="Output file path (For single file input only)")
    parser.add_argument('-d', '--directory', default=".",
                        help="Export to an existing directory. Default in current directory.")
    # parser.add_argument('-f', '--format',
    #                     help="""Default value depends on --type argument.
    #                     HangSeng: 11s58s35s25s24s
    #                     MPower: 11s12s78s46s
    #                     This is the number of bytes expected for each field. For
    #                     example, for HangSeng, we expect Date(11), Title(58),
    #                     Deposit(35), Withdraw(25), and Balance(24).
    #                     Default value is set based on my statements. If it
    #                     doesn't work as expected, use `--verbose` option and
    #                     adjust format string until an aligned table is printed
    #                     out.""")
    parser.add_argument('-t', '--type',
                        help="""Type of PDF.

                        Available types: HangSeng, MPower, DBS.""")
    parser.add_argument('-v', '--verbose', default=False, action="store_true",
                        help="More details.")
    parser.add_argument('file', nargs='+', help='One or more PDF eStatements to process.')

    args = parser.parse_args()
    if args.output and len(args.file) > 1:
        sys.exit("Output option can only be set with single file input. To export multiple files, use -d option.")

    for stmt in args.file:
        print("Processing: {}".format(stmt))
        if args.type.lower() == 'hangseng':
            allrecords = HangSengSavingsImporter("Dummy:Account:Name", "Dummy", debug=args.verbose).extract(_FileMemo(stmt))
        elif args.type.lower() == 'mpower':
            allrecords = MPowerMasterImporter("Dummy:Account:Name", "Dummy", debug=args.verbose).extract(_FileMemo(stmt))
        elif args.type.lower() == 'dbs':
            allrecords = DBSImporter("Dummy:Account:Name", "Dummy", debug=args.verbose).extract(_FileMemo(stmt))
        output_path = args.output if args.output else path.join(args.directory, path.splitext(path.basename(stmt))[0] + ".csv")
        print("Exporting to {}".format(output_path))

        with open(output_path, mode='w') as csv_file:
            csv_writer = csv.writer(csv_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            if args.type.lower() == 'hangseng':
                csv_writer.writerow(["date", "title", "amount"])
                for txn in allrecords:
                    csv_writer.writerow([txn.date.isoformat(), txn.payee, txn.postings[0].units.number])
            elif args.type.lower() == 'mpower':
                csv_writer.writerow(["trans_date", "post_date", "activity", "amount"])
                for txn in allrecords:
                    csv_writer.writerow([txn.meta['txn_date'], txn.date.isoformat(), txn.narration, txn.postings[0].units.number])
            elif args.type.lower() == 'dbs':
                csv_writer.writerow(["trans_date", "post_date", "description", "amount"])
                for txn in allrecords:
                    csv_writer.writerow([txn.meta['txn_date'], txn.date.isoformat(), txn.narration, txn.postings[0].units.number])

    return 0


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
import platform
from os import path

with open("README.md", "r") as fh:
    long_description = fh.read()

# Explicitly list the scripts to install.
binaries = [
    ('beancount-hangseng-csv', 'beancount_hangseng.scripts.csv')
]

setup_extra_kwargs = {}
if platform.system() == 'Windows':
    setup_extra_kwargs.update(entry_points={
        'console_scripts': [
            '{} = {}:main'.format(binary, module)
            for binary, module in binaries]
    })
else:
    setup_extra_kwargs.update(scripts=[
        path.join('bin', binary)
        for binary, _ in binaries])

setup(
    name="beancount-hangseng",
    version="0.4.0",
    author="Cheong Yiu Fung",
    author_email="mail@yiufung.net",
    description="Parse Hang Seng eStatements to Beancount/CSV",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yiufung/beancount-hangseng",
    install_requires=['beancount'],
    # pacakges=find_packages(),
    packages=[
        'beancount_hangseng',
        'beancount_hangseng.scripts'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Office/Business :: Financial :: Accounting",
        "Topic :: Office/Business :: Financial :: Spreadsheet",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    **setup_extra_kwargs
)
